import threading
from collections.abc import Iterator
from contextlib import AbstractContextManager, contextmanager
from datetime import timedelta
from typing import TYPE_CHECKING

import werkzeug.serving

if TYPE_CHECKING:
    from _typeshed.wsgi import WSGIApplication

from ._app import app

assert __package__


def run_server_in_thread(
    port: int = 0,
    *,
    require_client_registration: bool = False,
    require_nonce: bool = False,
    issue_refresh_token: bool = True,
    access_token_max_age: timedelta = timedelta(hours=1),
) -> AbstractContextManager[werkzeug.serving.BaseWSGIServer]:
    """Run a OIDC provider server on a background thread.

    The server is stopped when the context ends.

    See `app <oidc_provider_mock.app>` for documentation of parameters.

    >>> with run_server_in_thread(port=25432) as server:
    ...     print(f"Server listening at http://localhost:{server.server_port}")
    Server listening at http://localhost:25432

    """
    return _threaded_server(
        port=port,
        app=app(
            require_client_registration=require_client_registration,
            require_nonce=require_nonce,
            issue_refresh_token=issue_refresh_token,
            access_token_max_age=access_token_max_age,
        ),
    )


@contextmanager
def _threaded_server(
    app: "WSGIApplication",
    *,
    host: str = "localhost",
    port: int = 0,
    poll_interval: float = 0.1,
) -> Iterator[werkzeug.serving.ThreadedWSGIServer]:
    server = werkzeug.serving.make_server(
        host,
        port,
        app,
        threaded=True,
    )

    assert isinstance(server, werkzeug.serving.ThreadedWSGIServer)

    def run():
        try:
            server.serve_forever(poll_interval)
        finally:
            server.server_close()

    server_thread = threading.Thread(target=run)
    server_thread.start()

    try:
        yield server

    finally:
        shutdown_thread = threading.Thread(target=server.shutdown)
        shutdown_thread.start()
        shutdown_thread.join(1)
        if shutdown_thread.is_alive():
            raise TimeoutError("Server failed to shut down in time")

        server_thread.join(0.5)
        if server_thread.is_alive():
            raise TimeoutError("Server thread timed out")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
from pathlib import Path
from typing import Any

import docutils.nodes
import sphinx.addnodes
from sphinx.application import Sphinx
from sphinx.builders import Builder
from sphinx.domains import Domain
from sphinx.environment import BuildEnvironment
from typing_extensions import override

# Allow autodoc to import modules
sys.path.insert(0, str(Path("..", "src").resolve()))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "oidc-provider-mock"
copyright = "© 2025- Thomas Scholtes"
author = "Thomas Scholtes"

github_url = f"https://github.com/geigerzaehler/{project}"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

default_role = "any"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "myst_parser",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "flask": ("https://flask.palletsprojects.com/en/stable/", None),
}

autodoc_typehints = "both"
autodoc_typehints_description_target = "documented"

myst_heading_anchors = 2

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


templates_path = ["_templates"]
html_title = "OIDC Provider Mock"
html_theme = "shibuya"
html_static_path = ["_static"]
html_css_files = ["custom.css"]
html_theme_options = {"github_url": f"https://github.com/geigerzaehler/{project}"}

github_ref = os.getenv("READTHEDOCS_GIT_IDENTIFIER", "main")
github_src_url = f"{github_url}/blob/{github_ref}"


class GithubSourceDomain(Domain):
    @override
    def resolve_any_xref(
        self,
        env: BuildEnvironment,
        fromdocname: str,
        builder: Builder,
        target: str,
        node: sphinx.addnodes.pending_xref,
        contnode: docutils.nodes.Element,
    ) -> list[tuple[str, Any]]:
        if not Path(__file__).parent.parent.joinpath(target).exists():
            return []

        reference = docutils.nodes.reference(
            refuri=f"{github_src_url}/{target}",
            internal=False,
        )
        reference.append(contnode)

        return [("github:link", reference)]


def setup(app: Sphinx):
    app.add_domain(GithubSourceDomain)



# pyright: reportPrivateUsage=none

import dataclasses
import logging
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import timedelta
from typing import Any, TypeVar, cast

import flask
import pytest
import typeguard
import werkzeug.serving
from playwright.sync_api import Page

typeguard.install_import_hook("oidc_provider_mock")
import oidc_provider_mock  # noqa: E402
import oidc_provider_mock._server  # noqa: E402
from oidc_provider_mock._app import Config  # noqa: E402


@pytest.fixture
def app(request: pytest.FixtureRequest):
    node = cast("Any", request.node)  # pyright: ignore
    marker = node.get_closest_marker("provider_config")
    if marker:
        config = marker.args[0]
        app = oidc_provider_mock.app(**dataclasses.asdict(config))
    else:
        app = oidc_provider_mock.app()

    # When using the `client` fixture, we include the port so that authlib does
    # not complain about insecure URLs.
    app.config["SERVER_NAME"] = "localhost:54321"
    return app


_C = TypeVar("_C", bound=Callable[..., None])


def use_provider_config(
    *,
    require_client_registration: bool = False,
    require_nonce: bool = False,
    issue_refresh_token: bool = True,
    access_token_max_age: timedelta = timedelta(hours=1),
) -> Callable[[_C], _C]:
    """Set configuration for the app under test."""

    return pytest.mark.provider_config(
        Config(
            require_client_registration=require_client_registration,
            require_nonce=require_nonce,
            issue_refresh_token=issue_refresh_token,
            access_token_max_age=access_token_max_age,
        ),
    )


@pytest.fixture(autouse=True)
def setup_logging():
    logging.getLogger("oidc_provider_mock").setLevel(logging.DEBUG)


@pytest.fixture
def oidc_server(app: flask.Flask) -> Iterator[str]:
    with run_server(app) as server:
        yield server.url()


@pytest.fixture
def page(page: Page):
    page.set_default_navigation_timeout(3000)
    page.set_default_timeout(3000)
    return page


@dataclass
class TestServer:
    app: flask.Flask
    server: werkzeug.serving.BaseWSGIServer

    def url(self, path: str = ""):
        path = path.lstrip("/")
        return f"http://localhost:{self.server.server_port}/{path}"


@contextmanager
def run_server(app: flask.Flask) -> Iterator[TestServer]:
    with oidc_provider_mock._server._threaded_server(app, poll_interval=0.01) as server:
        app.config["SERVER_NAME"] = f"localhost:{server.server_port}"
        yield TestServer(app, server)



import logging
import secrets
import textwrap
import warnings
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from http import HTTPStatus
from typing import TypeVar, cast
from uuid import uuid4

import authlib.deprecate
import authlib.integrations.flask_oauth2 as flask_oauth2
import authlib.oauth2.rfc6749
import authlib.oauth2.rfc6749.errors
import authlib.oauth2.rfc6750
import authlib.oidc.core
import flask
import flask.typing
import pydantic
import werkzeug.debug
import werkzeug.exceptions
import werkzeug.local
from authlib import jose
from authlib.integrations.flask_oauth2.authorization_server import FlaskOAuth2Request
from authlib.oauth2 import OAuth2Error, OAuth2Request
from typing_extensions import Never, override

from . import _client
from ._storage import (
    AccessToken,
    AuthorizationCode,
    Client,
    ClientAllowAny,
    ClientAuthMethod,
    RefreshToken,
    Storage,
    User,
    storage,
)

assert __package__
_logger = logging.getLogger(__package__)

_JWS_ALG = "RS256"


class TokenValidator(authlib.oauth2.rfc6750.BearerTokenValidator):
    @override
    def authenticate_token(self, token_string: str):  # pyright: ignore[reportIncompatibleMethodOverride]
        token = storage.get_access_token(token_string)
        if not token:
            raise authlib.oauth2.rfc6749.AccessDeniedError()

        return token


class AuthorizationCodeGrant(authlib.oauth2.rfc6749.AuthorizationCodeGrant):
    @override
    def query_authorization_code(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, code: str, client: Client
    ) -> AuthorizationCode | None:
        auth_code = storage.get_authorization_code(code)
        if auth_code and auth_code.client_id == client.get_client_id():
            return auth_code

    @override
    def delete_authorization_code(self, authorization_code: AuthorizationCode):
        storage.remove_authorization_code(authorization_code.code)

    @override
    def authenticate_user(self, authorization_code: AuthorizationCode) -> User | None:
        return storage.get_user(authorization_code.user_id)

    @override
    def save_authorization_code(self, code: str, request: object):
        assert isinstance(request, OAuth2Request)
        assert isinstance(request.user, User)
        client = cast("Client", request.client)
        with warnings.catch_warnings():
            # Silence warnings for deprecated `OAuth2Request` properties.
            warnings.simplefilter("ignore", authlib.deprecate.AuthlibDeprecationWarning)
            assert isinstance(request.redirect_uri, str)  # pyright: ignore[reportDeprecated]
            storage.store_authorization_code(
                AuthorizationCode(
                    code=code,
                    user_id=request.user.sub,
                    client_id=client.get_client_id(),
                    redirect_uri=request.redirect_uri,  # pyright: ignore[reportDeprecated]
                    scope=request.scope,  # pyright: ignore[reportDeprecated]
                    nonce=request.data.get("nonce"),  # pyright: ignore[reportDeprecated]
                )
            )


class OpenIDCode(authlib.oidc.core.OpenIDCode):
    def __init__(self, require_nonce: bool, token_max_age: timedelta):
        super().__init__(require_nonce)
        self._token_max_mage = token_max_age

    @override
    def exists_nonce(self, nonce: str, request: OAuth2Request) -> bool:
        return storage.exists_nonce(nonce)

    @override
    def get_jwt_config(self, grant: authlib.oauth2.rfc6749.BaseGrant):  # pyright: ignore[reportUnknownParameterType]
        return {
            "key": storage.jwk.as_dict(is_private=True),  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
            "alg": _JWS_ALG,
            "exp": int(self._token_max_mage.total_seconds()),
            "iss": flask.request.host_url.rstrip("/"),
        }

    @override
    def generate_user_info(self, user: User, scope: str):  # pyright: ignore[reportIncompatibleMethodOverride]
        return _user_claims_for_scope(user, scope)


class RefreshTokenGrant(authlib.oauth2.rfc6749.RefreshTokenGrant):
    @override
    def authenticate_refresh_token(self, refresh_token: str):
        token = storage.get_refresh_token(refresh_token)
        if not token:
            raise authlib.oauth2.rfc6749.InvalidGrantError("invalid refresh token")

        return token

    def authenticate_user(self, refresh_token: RefreshToken):
        return storage.get_user(refresh_token.user_id)

    def revoke_old_credential(self, refresh_token: authlib.oauth2.rfc6749.TokenMixin):
        assert isinstance(refresh_token, RefreshToken)
        storage.remove_access_token(refresh_token.access_token)


def _user_claims_for_scope(user: User, scope: str) -> dict[str, object]:
    scopes = scope.split(" ")
    allowed_standard_claims_for_scope = {
        claim for scope in scopes for claim in _SCOPES_TO_CLAIMS.get(scope, [])
    }

    return {
        **{
            name: value
            for name, value in user.claims.items()
            if name not in _STANDARD_CLAIMS or name in allowed_standard_claims_for_scope
        },
        "sub": user.sub,
    }


# https://openid.net/specs/openid-connect-core-1_0.html#ScopeClaims
_SCOPES_TO_CLAIMS: dict[str, Sequence[str]] = {
    "profile": [
        "name",
        "family_name",
        "given_name",
        "middle_name",
        "nickname",
        "preferred_username",
        "profile",
        "picture",
        "website",
        "gender",
        "birthdate",
        "zoneinfo",
        "locale",
        "updated_at",
    ],
    "email": ["email", "email_verified"],
    "address": ["address"],
    "phone": ["phone_number", "phone_number_verified"],
}

_STANDARD_CLAIMS = {claim for claims in _SCOPES_TO_CLAIMS.values() for claim in claims}


require_oauth = flask_oauth2.ResourceProtector()

authorization = cast(
    "flask_oauth2.AuthorizationServer",
    werkzeug.local.LocalProxy(lambda: flask.g._authlib_authorization_server),
)

blueprint = flask.Blueprint("oidc-provider-mock", __name__)


@dataclass(kw_only=True, frozen=True)
class Config:
    require_client_registration: bool = False
    require_nonce: bool = False
    issue_refresh_token: bool = True
    access_token_max_age: timedelta = timedelta(hours=1)


@blueprint.record
def setup(setup_state: flask.blueprints.BlueprintSetupState):
    assert isinstance(setup_state.app, flask.Flask)

    config = setup_state.options["config"]
    assert isinstance(config, Config)

    setup_state.app.config["OAUTH2_TOKEN_EXPIRES_IN"] = {
        "authorization_code": int(config.access_token_max_age.total_seconds()),
    }

    setup_state.app.config["OAUTH2_REFRESH_TOKEN_GENERATOR"] = (
        config.issue_refresh_token
    )

    authorization = flask_oauth2.AuthorizationServer()
    storage = Storage()

    @setup_state.app.before_request
    def set_globals():
        flask.g.oidc_provider_mock_storage = storage
        flask.g.oidc_provider_mock_config = config
        flask.g._authlib_authorization_server = authorization

    def query_client(id: str) -> Client | None:
        client = storage.get_client(id)
        if not client and not config.require_client_registration:
            client = Client(
                id=id,
                secret=ClientAllowAny(),
                redirect_uris=ClientAllowAny(),
                allowed_scopes=Client.SCOPES_SUPPORTED,
                token_endpoint_auth_method=ClientAllowAny(),
            )

        return client

    def save_token(token: dict[str, object], request: OAuth2Request):
        assert token["token_type"] == "Bearer"
        assert isinstance(token["access_token"], str)
        assert isinstance(token["expires_in"], int)
        assert isinstance(request.user, User)
        scope = token.get("scope", "")
        assert isinstance(scope, str)

        storage.store_access_token(
            AccessToken(
                token=token["access_token"],
                user_id=request.user.sub,
                # request.scope may actually be None
                scope=scope,
                expires_at=datetime.now(timezone.utc)
                + timedelta(seconds=token["expires_in"]),
            )
        )

        if "refresh_token" in token:
            assert isinstance(token["refresh_token"], str)
            assert isinstance(request.client, Client)

            storage.store_refresh_token(
                RefreshToken(
                    access_token=token["access_token"],
                    token=token["refresh_token"],
                    user_id=request.user.sub,
                    scope=scope,
                    expires_at=datetime.now(timezone.utc)
                    + timedelta(seconds=token["expires_in"]),
                    client_id=request.client.id,
                )
            )

    authorization.init_app(  # type: ignore
        setup_state.app,
        query_client=query_client,
        save_token=save_token,
    )

    authorization.register_grant(
        AuthorizationCodeGrant,
        [
            OpenIDCode(
                require_nonce=config.require_nonce,
                token_max_age=config.access_token_max_age,
            )
        ],
    )

    authorization.register_grant(RefreshTokenGrant)


@blueprint.record_once
def setup_once(setup_state: flask.blueprints.BlueprintSetupState):
    require_oauth.register_token_validator(TokenValidator())  # pyright: ignore[reportUnknownMemberType]


def app(
    *,
    require_client_registration: bool = False,
    require_nonce: bool = False,
    issue_refresh_token: bool = True,
    access_token_max_age: timedelta = timedelta(hours=1),
) -> flask.Flask:
    """Create a Flask app running the OpenID provider.

    Call ``app().run()`` (see `flask.Flask.run`) to start the server.

    See ``init_app`` for documentation of parameters
    """
    app = flask.Flask(__name__)

    init_app(
        app,
        require_client_registration=require_client_registration,
        require_nonce=require_nonce,
        issue_refresh_token=issue_refresh_token,
        access_token_max_age=access_token_max_age,
    )
    app.secret_key = secrets.token_bytes(16)
    if isinstance(app.json, flask.json.provider.DefaultJSONProvider):
        # Make it easier to debug responses
        app.json.compact = False
    return app


def init_app(
    app: flask.Flask,
    *,
    require_client_registration: bool = False,
    require_nonce: bool = False,
    issue_refresh_token: bool = True,
    access_token_max_age: timedelta = timedelta(hours=1),
):
    """Add the OpenID provider and its endpoints to the flask ``app``.

    :param require_client_registration: If false (the default) any client ID and
        secret can be used to authenticate with the token endpoint. If true,
        clients have to be registered using the `OAuth 2.0 Dynamic Client
        Registration Protocol <https://datatracker.ietf.org/doc/html/rfc7591>`_.
    :param require_nonce: If true, the authorization request must include the
        `nonce parameter`_ to prevent replay attacks. If the parameter is not
        provided the authorization request will fail.
    :param issue_refresh_token: If true (the default), the token endpoint response
        will include a refresh token.
    :param access_token_max_age: Max age of access and ID token after which it expires.

    .. _nonce parameter: https://openid.net/specs/openid-connect-core-1_0.html#AuthRequest
    """

    app.register_blueprint(
        blueprint,
        config=Config(
            require_client_registration=require_client_registration,
            require_nonce=require_nonce,
            issue_refresh_token=issue_refresh_token,
            access_token_max_age=access_token_max_age,
        ),
    )

    app.register_blueprint(_client.blueprint)

    app.debug = True
    app.wsgi_app = werkzeug.debug.DebuggedApplication(app.wsgi_app)
    app.wsgi_app.trusted_hosts.append("localhost")

    return app


@blueprint.get("/")
def home():
    return flask.render_template("index.html")


@blueprint.get("/.well-known/openid-configuration")
def openid_config():
    def url_for(fn: Callable[..., object]) -> str:
        return flask.url_for(f".{fn.__name__}", _external=True)

    # See https://openid.net/specs/openid-connect-discovery-1_0.html#ProviderMetadata
    # for information about the fields.
    return flask.jsonify({
        "issuer": flask.request.host_url.rstrip("/"),
        "authorization_endpoint": url_for(authorize),
        "token_endpoint": url_for(issue_token),
        "userinfo_endpoint": url_for(userinfo),
        "registration_endpoint": url_for(register_client),
        "jwks_uri": url_for(jwks),
        "response_types_supported": Client.RESPONSE_TYPES_SUPPORTED,
        "response_modes_supported": ["query"],
        "grant_types_supported": Client.GRANT_TYPES_SUPPORTED,
        "scopes_supported": Client.SCOPES_SUPPORTED,
        "id_token_signing_alg_values_supported": [_JWS_ALG],
    })


@blueprint.get("/jwks")
def jwks():
    return flask.jsonify(
        jose.KeySet((storage.jwk,)).as_dict(),  # pyright: ignore[reportUnknownMemberType]
    )


class RegisterClientBody(pydantic.BaseModel):
    redirect_uris: Sequence[pydantic.HttpUrl]
    token_endpoint_auth_method: ClientAuthMethod = "client_secret_basic"
    scope: str | None = None


@blueprint.post("/oauth2/clients")
def register_client():
    body = _validate_body(flask.request, RegisterClientBody)

    client = Client(
        id=str(uuid4()),
        secret=secrets.token_urlsafe(16),
        redirect_uris=[str(uri) for uri in body.redirect_uris],
        allowed_scopes=body.scope or Client.SCOPES_SUPPORTED,
        token_endpoint_auth_method=body.token_endpoint_auth_method,
    )

    storage.store_client(client)
    return flask.jsonify({
        "client_id": client.id,
        "client_secret": client.secret,
        "redirect_uris": client.redirect_uris,
        "token_endpoint_auth_method": body.token_endpoint_auth_method,
        "grant_types": Client.GRANT_TYPES_SUPPORTED,
        "response_types": Client.RESPONSE_TYPES_SUPPORTED,
    }), HTTPStatus.CREATED


@blueprint.route("/oauth2/authorize", methods=["GET", "POST"])
def authorize() -> flask.typing.ResponseReturnValue:
    request = FlaskOAuth2Request(flask.request)
    try:
        grant, redirect_uri = _validate_auth_request_client_params(flask.request)
        assert isinstance(grant.client, Client)  # pyright: ignore[reportUnknownMemberType]
    except _AuthorizationValidationException as exc:
        _logger.warning(f"invalid authorization request: {exc.description}")
        raise

    if flask.request.method == "GET":
        return flask.render_template(
            "authorization_form.html",
            redirect_uri=redirect_uri,
            client_id=grant.client.id,
            recent_subjects=storage.get_recent_subjects(),
        )
    else:
        if flask.request.form.get("action") == "deny":
            return authorization.handle_response(  # pyright: ignore[reportUnknownMemberType]
                *authlib.oauth2.rfc6749.AccessDeniedError(  # pyright: ignore[reportUnknownArgumentType]
                    redirect_uri=flask.request.args["redirect_uri"]
                )()
            )

        sub = flask.request.form.get("sub")
        if sub is None:
            return flask.render_template(
                "authorization_form.html",
                redirect_uri=redirect_uri,
                client_id=grant.client.id,
                recent_subjects=storage.get_recent_subjects(),
                sub_missing=True,
            )

        user = storage.get_user(sub)
        if not user:
            user = User(sub=sub, claims={"email": sub})
            storage.store_user(user)

        try:
            response = grant.create_authorization_response(redirect_uri, user)  # pyright: ignore
            _logger.info(
                "issued authorization code",
                extra=({"client": grant.client, "user": user}),
            )
            storage.record_subject(sub)
            return authorization.handle_response(*response)  # pyright: ignore
        except authlib.oauth2.OAuth2Error as error:
            _logger.warning("invalid authorization request", exc_info=True)
            return authorization.handle_error_response(request, error)  # pyright: ignore


def _validate_auth_request_client_params(
    flask_request: flask.Request,
) -> tuple[AuthorizationCodeGrant | RefreshTokenGrant, str]:
    """Validate query parameters sent by the client to the authorization endpoint.

    Raises ``_AuthorizationValidationException`` if validation fails which results
    in an appropriate 400 response.
    """

    request = FlaskOAuth2Request(flask_request)

    try:
        grant = authorization.get_consent_grant()  # type: ignore
        assert isinstance(grant, AuthorizationCodeGrant)
        redirect_uri = grant.validate_authorization_request()
    except authlib.oauth2.rfc6749.InvalidClientError as e:
        raise _AuthorizationValidationException(
            authlib.oauth2.rfc6749.InvalidClientError.error,
            e.description,
        ) from e
    except authlib.oauth2.rfc6749.UnsupportedResponseTypeError as e:
        raise _AuthorizationValidationException(
            e.error,
            f"OAuth response_type {e.response_type} is not supported",
        ) from e
    except authlib.oauth2.rfc6749.InvalidRequestError as e:
        description = e.description
        # FIXME: this is a brittle way of determining what the error is but
        # authlib does not raise a dedicated error in this case.
        if description == "Redirect URI foo is not supported by client.":
            raise _AuthorizationValidationException(
                authlib.oauth2.rfc6749.InvalidClientError.error,
                description,
            ) from e
        else:
            raise werkzeug.exceptions.HTTPException(
                response=flask.make_response(
                    authorization.handle_error_response(request, e)
                )
            ) from e
    except authlib.oauth2.OAuth2Error as e:
        raise werkzeug.exceptions.HTTPException(
            response=flask.make_response(
                authorization.handle_error_response(request, e)
            )
        ) from e

    return grant, redirect_uri


class _AuthorizationValidationException(werkzeug.exceptions.HTTPException):
    error: str

    def __init__(self, error: str, description: str):
        self.error = error
        self.description = description
        response = flask.make_response(
            flask.render_template("error.html", name=error, description=description),
            HTTPStatus.BAD_REQUEST,
        )
        super().__init__(response=response)
        self.code = HTTPStatus.BAD_REQUEST


@blueprint.post("/oauth2/token")
def issue_token() -> flask.typing.ResponseReturnValue:
    request = FlaskOAuth2Request(flask.request)
    try:
        grant = authorization.get_token_grant(request)
    except authlib.oauth2.rfc6749.UnsupportedGrantTypeError as error:
        _logger.warning(
            "unsupported grant type for issuing token",
            extra={"grant_type": error.grant_type},
        )
        return authorization.handle_error_response(request, error)  # type: ignore

    assert isinstance(grant, AuthorizationCodeGrant | RefreshTokenGrant)

    try:
        grant.validate_token_request()
        args = grant.create_token_response()
        return authorization.handle_response(*args)  # type: ignore
    except OAuth2Error as error:
        if error.error:
            _logger.warning(
                f"token endpoint error {error.error}",
                extra={"description": error.description},
            )
        else:
            _logger.warning("error while issuing token", exc_info=error)
        return authorization.handle_error_response(request, error)  # type: ignore


@blueprint.route("/userinfo", methods=["GET", "POST"])
@require_oauth()
def userinfo():
    access_token = flask_oauth2.current_token
    assert isinstance(access_token, AccessToken)
    return flask.jsonify(
        _user_claims_for_scope(access_token.get_user(), access_token.scope)
    )


SetUserBody = pydantic.RootModel[dict[str, object]]


@blueprint.put("/users/<sub>")
def set_user(sub: str):
    body = _validate_body(flask.request, SetUserBody)
    storage.store_user(User(sub=sub, claims=body.root))
    return "", HTTPStatus.NO_CONTENT


@blueprint.post("/users/<sub>/revoke-tokens")
def revoke_user_tokens(sub: str):
    for access_token in storage.access_tokens():
        if access_token.user_id == sub:
            storage.remove_access_token(access_token.token)
    for refresh_token in storage.refresh_tokens():
        if refresh_token.user_id == sub:
            storage.remove_refresh_token(refresh_token.token)
    return "", HTTPStatus.NO_CONTENT


class InsecureTransportError(Exception):
    def __init__(self):
        super().__init__(
            "OAuth 2 requires https. Set the environment variable"
            "`AUTHLIB_INSECURE_TRANSPORT=1` to disable this check"
        )


def _insecure_transport_error_handler(
    error: authlib.oauth2.rfc6749.errors.InsecureTransportError,
) -> Never:
    raise InsecureTransportError() from error


blueprint.register_error_handler(
    authlib.oauth2.rfc6749.errors.InsecureTransportError,
    _insecure_transport_error_handler,
)

_Model = TypeVar("_Model", bound=pydantic.BaseModel)


def _validate_body(request: flask.Request, model: type[_Model]) -> _Model:
    try:
        return model.model_validate(request.json, strict=True)
    except pydantic.ValidationError as error:
        _logger.info(
            f"invalid request body {request.method} {request.url}\n{textwrap.indent(str(error), '  ')}",
            extra={
                "_msg": "invalid request body",
                "method": request.method,
                "url": request.url,
                "error": error,
            },
        )

        # TODO: support content type negotiation with html and json
        msg = "Invalid body:\n"
        for detail in error.errors():
            loc = detail.get("loc")
            if loc:
                msg += f"- {_pydantic_loc_to_path(loc)}:"
            msg += f" {detail.get('msg')}\n"

        raise werkzeug.exceptions.HTTPException(
            response=flask.make_response(
                msg,
                HTTPStatus.BAD_REQUEST,
                {"content-type": "text/plain; charset=utf-8"},
            )
        ) from error


def _pydantic_loc_to_path(loc: tuple[str | int, ...]) -> str:
    path = ""
    for i, x in enumerate(loc):
        match x:
            case str():
                if i > 0:
                    path += "."
                path += x
            case int():
                path += f"[{x}]"
    return path



#!/usr/bin/env -S uv run

import re
from collections.abc import Iterable
from pathlib import Path


def main():
    output_path = Path("README.md")
    template_path = Path("README.md.tpl")
    snippets = _extract_snippets((Path("examples/flask_oidc_example.py"),))

    def replace(match: re.Match[str]) -> str:
        name = match.group(1)
        return snippets.pop(name)

    template = template_path.read_text()

    with output_path.open("w") as output:
        output.write("<!-- THIS FILE HAS BEEN GENERATED FROM REAMDE.md.tpl -->\n")
        output.write(re.sub(r"SNIPPET (\S+)", replace, template))

    if snippets:
        raise RuntimeError(f"Unused snippets {','.join(snippets.keys())}")


def _extract_snippets(paths: Iterable[Path]):
    snippets = dict[str, str]()
    for path in paths:
        content = path.read_text()
        matches = re.findall(
            r"# SNIPPET (\S+)\n+(.*?)\n+# SNIPPET END", content, re.DOTALL
        )
        for name, code in matches:
            if name in snippets:
                raise ValueError(f"Duplicate snippet name {name}")
            snippets[name] = code

    return snippets


main()



from collections import deque
from collections.abc import Collection, Iterable, Sequence
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import ClassVar, Literal, TypeAlias, cast

import authlib.oauth2.rfc6749
import authlib.oidc.core
import flask
import werkzeug.local
from authlib import jose
from typing_extensions import override


class ClientAllowAny:
    """Special value for client fields that skips validation of the field."""

    def __repr__(self) -> str:
        return str(type(self).__name__)


ClientAuthMethod: TypeAlias = (
    Literal["none"] | Literal["client_secret_basic"] | Literal["client_secret_post"]
)


@dataclass(kw_only=True, frozen=True)
class Client(authlib.oauth2.rfc6749.ClientMixin):
    id: str
    secret: str | ClientAllowAny
    redirect_uris: Sequence[str] | ClientAllowAny
    allowed_scopes: Sequence[str]
    token_endpoint_auth_method: ClientAuthMethod | ClientAllowAny

    """Wrap ``Client`` to implement authlib’s client protocol."""

    RESPONSE_TYPES_SUPPORTED: ClassVar[tuple[str, ...]] = ("code",)
    GRANT_TYPES_SUPPORTED: ClassVar[tuple[str, ...]] = (
        "authorization_code",
        "refresh_token",
    )
    SCOPES_SUPPORTED: ClassVar[tuple[str, ...]] = (
        "openid",
        "profile",
        "email",
        "address",
        "phone",
    )

    @override
    def get_client_id(self):
        return self.id

    @override
    def get_default_redirect_uri(self) -> str | None:  # pyright: ignore[reportIncompatibleMethodOverride]
        if isinstance(self.redirect_uris, ClientAllowAny):
            return None

        return self.redirect_uris[0]

    @override
    def get_allowed_scope(self, scope: Collection[str] | str) -> str:
        if isinstance(scope, str):
            scopes = scope.split()
        else:
            scopes = scope
        return " ".join(s for s in scopes if s in self.allowed_scopes)

    @override
    def check_redirect_uri(self, redirect_uri: str) -> bool:
        if isinstance(self.redirect_uris, ClientAllowAny):
            return True

        return redirect_uri in self.redirect_uris

    @override
    def check_client_secret(self, client_secret: str) -> bool:
        if isinstance(self.secret, ClientAllowAny):
            return True

        return client_secret == self.secret

    @override
    def check_endpoint_auth_method(self, method: str, endpoint: object):
        if isinstance(self.token_endpoint_auth_method, ClientAllowAny):
            return True

        return method == self.token_endpoint_auth_method

    @override
    def check_grant_type(self, grant_type: str):
        return grant_type in self.GRANT_TYPES_SUPPORTED

    @override
    def check_response_type(self, response_type: str):
        return response_type in self.RESPONSE_TYPES_SUPPORTED


@dataclass(kw_only=True, frozen=True)
class User:
    sub: str
    claims: dict[str, object] = field(default_factory=dict[str, object])


@dataclass(kw_only=True, frozen=True)
class AuthorizationCode(authlib.oidc.core.AuthorizationCodeMixin):
    code: str
    client_id: str
    redirect_uri: str
    user_id: str
    scope: str
    nonce: str | None

    # Implement AuthorizationCodeMixin

    @override
    def get_redirect_uri(self):
        return self.redirect_uri

    @override
    def get_scope(self):
        return self.scope

    @override
    def get_nonce(self) -> str | None:
        return self.nonce

    @override
    def get_auth_time(self) -> int | None:
        return None


@dataclass(kw_only=True, frozen=True)
class AccessToken(authlib.oauth2.rfc6749.TokenMixin):
    token: str
    user_id: str
    scope: str
    expires_at: datetime

    def get_user(self) -> User:
        user = storage.get_user(self.user_id)
        if user is None:
            raise RuntimeError(f"Missing user {self.user_id} for access toke")
        return user

    # Implement `TokenMixin`

    @override
    def check_client(self, client: Client) -> bool:
        # Required only for revocation and refresh token endpoints.
        raise NotImplementedError()

    @override
    def is_expired(self):
        return datetime.now(timezone.utc) >= self.expires_at

    @override
    def is_revoked(self):
        return False

    @override
    def get_scope(self) -> str:
        return self.scope


@dataclass(kw_only=True, frozen=True)
class RefreshToken(AccessToken):
    client_id: str
    access_token: str

    @override
    def check_client(self, client: Client) -> bool:
        return self.client_id == client.id


class Storage:
    jwk: jose.RSAKey

    _clients: dict[str, Client]
    _users: dict[str, User]
    _authorization_codes: dict[str, AuthorizationCode]
    _access_tokens: dict[str, AccessToken]
    _refresh_tokens: dict[str, RefreshToken]
    _nonces: set[str]
    _recent_subjects: deque[str]

    def __init__(self) -> None:
        self.jwk = jose.RSAKey.generate_key(is_private=True)  # pyright: ignore[reportUnknownMemberType]
        self._clients = {}
        self._users = {}
        self._authorization_codes = {}
        self._access_tokens = {}
        self._refresh_tokens = {}
        self._nonces = set()
        self._recent_subjects = deque()

    # User

    def get_user(self, sub: str) -> User | None:
        return self._users.get(sub)

    def store_user(self, user: User):
        self._users[user.sub] = user

    def get_recent_subjects(self) -> Sequence[str]:
        """Get a sequence of the 20 most recently recorded subjects, starting with
        the most recent one.
        """
        return self._recent_subjects

    def record_subject(self, sub: str) -> None:
        try:
            self._recent_subjects.remove(sub)
        except ValueError:
            pass

        self._recent_subjects.appendleft(sub)
        if len(self._recent_subjects) > 20:
            self._recent_subjects.pop()

    # AuthorizationCodes

    def get_authorization_code(self, code: str) -> AuthorizationCode | None:
        return self._authorization_codes.get(code)

    def store_authorization_code(self, code: AuthorizationCode):
        self._authorization_codes[code.code] = code

    def remove_authorization_code(self, code: str) -> AuthorizationCode | None:
        return self._authorization_codes.pop(code, None)

    # AccessTokens

    def get_access_token(self, token: str) -> AccessToken | None:
        return self._access_tokens.get(token)

    def store_access_token(self, access_token: AccessToken):
        self._access_tokens[access_token.token] = access_token

    def remove_access_token(self, access_token: str) -> AccessToken | None:
        return self._access_tokens.pop(access_token, None)

    def access_tokens(self) -> Iterable[AccessToken]:
        return list(self._access_tokens.values())

    # RefreshTokens

    def get_refresh_token(self, token: str) -> RefreshToken | None:
        return self._refresh_tokens.get(token)

    def store_refresh_token(self, refresh_token: RefreshToken):
        self._refresh_tokens[refresh_token.token] = refresh_token

    def remove_refresh_token(self, token: str) -> RefreshToken | None:
        return self._refresh_tokens.pop(token, None)

    def refresh_tokens(self) -> Iterable[RefreshToken]:
        return list(self._refresh_tokens.values())

    # Client

    def get_client(self, id: str) -> Client | None:
        return self._clients.get(id)

    def store_client(self, client: Client):
        self._clients[client.id] = client

    # Nonce

    def add_nonce(self, nonce: str):
        self._nonces.add(nonce)

    def exists_nonce(self, nonce: str) -> bool:
        return nonce in self._nonces


storage = cast(
    "Storage", werkzeug.local.LocalProxy(lambda: flask.g.oidc_provider_mock_storage)
)



import secrets
from collections.abc import Callable
from http import HTTPStatus
from typing import Any

import flask
import htpy as h

from ._client_lib import AuthorizationError, OidcClient

blueprint = flask.Blueprint("oidc-client", __name__)


def _url_for(fn: Callable[..., object], **kwargs: Any) -> str:
    return flask.url_for(f".{fn.__name__}", **kwargs)


_SESSION_KEY_NONCE = "oidc_id_token_nonce"
_SESSION_KEY_STATE = "oidc_authorization_state"
_SESSION_KEY_OIDC_CLAIMS = "oidc_id_token_claims"


@blueprint.get("/oidc/login")
def login():
    state = secrets.token_urlsafe(16)
    nonce = secrets.token_urlsafe(16)
    flask.session[_SESSION_KEY_STATE] = state
    flask.session[_SESSION_KEY_NONCE] = nonce
    return _render_page(
        h.h1["Authenticate with OpenID Connect"],
        h.p[
            f"Start authentication with OpenID Connect Provider {flask.request.root_url}"
        ],
        h.form(method="post")[h.button["Start"]],
    )


@blueprint.post("/oidc/login")
def login_post():
    client = _get_client()
    state = secrets.token_urlsafe(16)
    nonce = secrets.token_urlsafe(16)
    flask.session[_SESSION_KEY_STATE] = state
    flask.session[_SESSION_KEY_NONCE] = nonce
    return flask.redirect(client.authorization_url(state=state, nonce=nonce))


@blueprint.get("/oidc/authorized")
def authorized():
    client = _get_client()
    if _SESSION_KEY_STATE not in flask.session:
        return _render_page(
            h.h1["Authentication Error"],
            h.p[
                "No authentication session in progress. ",
                h.a(href=_url_for(login))["Restart authentication"],
            ],
        )

    try:
        token_data = client.fetch_token(
            flask.request.url,
            state=flask.session[_SESSION_KEY_STATE],
        )
    except AuthorizationError as e:
        return _render_page(
            h.hgroup[h.h1["Authentication Error"], h.p[e.error]],
            h.p[e.description],
        ), HTTPStatus.BAD_REQUEST

    flask.session[_SESSION_KEY_OIDC_CLAIMS] = token_data.claims
    return flask.redirect(_url_for(success))


@blueprint.get("/oidc/success")
def success():
    claims = flask.session[_SESSION_KEY_OIDC_CLAIMS]
    return _render_page(
        h.h1["Authentication Success"],
        h.p["You’re logged in as ", h.mark[claims["sub"]]],
    )


def _get_client() -> OidcClient:
    return OidcClient(
        id="my-client-id",
        secret="my-client-secret",
        redirect_uri=_url_for(authorized, _external=True),
        issuer=flask.request.root_url,
    )


def _render_page(*content: h.Node) -> str:
    return flask.render_template("_base.html", content=h.fragment[content])



import inspect
from collections.abc import Callable
from http import HTTPStatus

import flask.testing

import oidc_provider_mock
import oidc_provider_mock._app

from .conftest import use_provider_config


def test_userinfo_unauthorized(client: flask.testing.FlaskClient):
    response = client.get("/userinfo")
    assert response.status_code == HTTPStatus.UNAUTHORIZED
    assert response.www_authenticate.type == "bearer"
    assert response.json
    assert response.json["error"] == "missing_authorization"

    response = client.get("/userinfo", headers={"authorization": "foo"})
    assert response.status_code == HTTPStatus.UNAUTHORIZED
    assert response.json == {"error": "unsupported_token_type"}

    response = client.get("/userinfo", headers={"authorization": "Bearer foo"})
    # Should be `HTTPStatus.UNAUTHORIZED` but there is a bug in authlib
    assert response.status_code == HTTPStatus.BAD_REQUEST
    assert response.json
    assert response.json["error"] == "access_denied"


def test_consistent_kwargs():
    """Check that kwargs for configuring the provider are consistent across all APIs"""

    def kw_only_params(obj: Callable[..., object]):
        signature = inspect.signature(obj)

        return tuple(
            (k, v.annotation, v.default)
            for k, v in signature.parameters.items()
            if v.kind
            if v.kind == inspect.Parameter.KEYWORD_ONLY
        )

    expected_params = kw_only_params(oidc_provider_mock._app.Config)

    assert kw_only_params(oidc_provider_mock.init_app) == expected_params
    assert kw_only_params(oidc_provider_mock.run_server_in_thread) == expected_params
    assert kw_only_params(use_provider_config) == expected_params



from ._app import app, init_app
from ._server import run_server_in_thread

__all__ = [  # noqa: RUF022
    # Custom order, respected by API docs
    "init_app",
    "app",
    "run_server_in_thread",
]



import subprocess
import time

import httpx


def test_cli():
    with subprocess.Popen(
        ["oidc-provider-mock"],
        stdin=None,
        text=True,
    ) as process:
        try:
            base_url = "http://127.0.0.1:9400"
            response = None
            for _ in range(5):
                try:
                    response = httpx.get(f"{base_url}/.well-known/openid-configuration")
                except httpx.ConnectError:
                    time.sleep(0.5)

            assert response
            assert response.status_code == 200
            body = response.json()
            assert body["issuer"] == base_url
        finally:
            process.terminate()



# pyright: reportUnknownMemberType=none
"""Test OIDC login of a Flask app using
[Flask-OIDC](https://flask-oidc.readthedocs.io/en/stable/).
"""

from urllib.parse import quote

import httpx
import pytest
from freezegun import freeze_time
from playwright.sync_api import Page, expect

from examples import flask_oidc_example

from .conftest import TestServer, run_server, use_provider_config


@pytest.fixture
def relying_party(oidc_server: str):
    with run_server(flask_oidc_example.build_app(oidc_server)) as server:
        yield server


def test_refresh(relying_party: TestServer, page: Page, oidc_server: str):
    with freeze_time("1 Jan 2020", tick=True) as frozen_datetime:
        response = httpx.put(
            f"{oidc_server}/users/{quote('alice@example.com')}",
            json={"email": "alice@example.com", "name": "Alice"},
        )
        assert response.status_code == 204

        page.goto(relying_party.url("/login"))
        page.get_by_placeholder("sub").fill("alice@example.com")
        page.get_by_role("button", name="Authorize").click()

        expect(page.locator("body")).to_contain_text(
            "Welcome Alice (alice@example.com)"
        )

        frozen_datetime.tick(7200)
        page.reload()
        expect(page.locator("body")).to_contain_text(
            "Welcome Alice (alice@example.com)"
        )


@use_provider_config(issue_refresh_token=False)
def test_access_token_expired(relying_party: TestServer, oidc_server: str, page: Page):
    with freeze_time("1 Jan 2020", tick=True) as frozen_datetime:
        response = httpx.put(
            f"{oidc_server}/users/{quote('alice@example.com')}",
            json={"email": "alice@example.com", "name": "Alice"},
        )
        assert response.status_code == 204

        page.goto(relying_party.url("/login"))
        page.get_by_placeholder("sub").fill("alice@example.com")
        page.get_by_role("button", name="Authorize").click()

        expect(page.locator("body")).to_contain_text(
            "Welcome Alice (alice@example.com)"
        )

        frozen_datetime.tick(7200)
        page.reload()
        expect(page.locator("body")).to_contain_text("Not logged in")



from faker import Faker
from playwright.sync_api import Page, expect

faker = Faker()


# @use_provider_config(require_client_registration=True)
def test_auth_success(oidc_server: str, page: Page):
    subject = faker.email()

    page.goto(f"{oidc_server}/oidc/login")
    page.get_by_role("button", name="Start").click()
    expect(page.get_by_role("heading")).to_have_text("Authorize Client")
    page.get_by_placeholder("sub").fill(subject)
    page.get_by_role("button", name="Authorize").click()
    expect(page.locator("body")).to_contain_text(f"You’re logged in as {subject}")


def test_auth_deny(oidc_server: str, page: Page):
    page.goto(f"{oidc_server}/oidc/login")
    page.get_by_role("button", name="Start").click()
    page.get_by_role("button", name="Deny").click()
    expect(page.get_by_role("heading")).to_have_text("Authentication Error")
    expect(page.locator("body")).to_contain_text("access_denied")



"""Tests authentication flow only using the POST requests to the authorization
endpoint.
"""

import re
from datetime import timedelta
from http import HTTPStatus
from typing import Any

import flask.testing
import httpx
import pytest
from authlib.integrations.base_client import OAuthError
from faker import Faker
from freezegun import freeze_time

from oidc_provider_mock._client_lib import (
    AuthorizationError,
    AuthorizationServerError,
    OidcClient,
)

from .conftest import use_provider_config

faker = Faker()


@use_provider_config(require_client_registration=True)
def test_auth_success(oidc_server: str):
    """Authorization Code flow success with client registration"""

    subject = faker.email()
    state = faker.password()
    nonce = faker.password()
    redirect_uri = faker.uri(schemes=["https"])

    client = OidcClient.register(oidc_server, redirect_uri=redirect_uri)

    response = httpx.post(
        client.authorization_url(state=state, nonce=nonce),
        data={"sub": subject},
    )
    assert response.status_code == 302
    location = response.headers["location"]
    assert location.startswith(redirect_uri)
    token_data = client.fetch_token(location, state=state)

    assert token_data.claims["sub"] == subject
    assert token_data.claims["email"] == subject
    assert token_data.claims["nonce"] == nonce

    userinfo = client.fetch_userinfo(token=token_data.access_token)
    assert userinfo["sub"] == subject


def test_custom_claims(oidc_server: str):
    """Authenticate with additional claims in ID token and user info"""

    subject = faker.email()
    state = faker.password()

    httpx.put(
        f"{oidc_server}/users/{subject}",
        json={"custom": "CLAIM"},
    ).raise_for_status()

    client = _fake_client(issuer=oidc_server)

    response = httpx.post(
        client.authorization_url(state=state),
        data={"sub": subject},
    )

    token_data = client.fetch_token(response.headers["location"], state=state)
    assert token_data.claims["sub"] == subject
    assert token_data.claims["custom"] == "CLAIM"

    userinfo = client.fetch_userinfo(token=token_data.access_token)
    assert userinfo["sub"] == subject
    assert userinfo["custom"] == "CLAIM"


def test_include_all_claims(oidc_server: str):
    subject = faker.email()
    state = faker.password()
    claims: dict[str, Any] = {
        # profile scope
        "name": faker.name(),
        "website": faker.uri(),
        # email scope
        "email": faker.email(),
        # address scope
        "address": {
            "formatted": faker.address(),
        },
        # phone scope
        "phone": faker.phone_number(),
    }

    httpx.put(f"{oidc_server}/users/{subject}", json=claims).raise_for_status()

    client = _fake_client(
        issuer=oidc_server,
        scope="openid profile email address phone",
    )

    response = httpx.post(
        client.authorization_url(state=state),
        data={"sub": subject},
    )

    token_data = client.fetch_token(response.headers["location"], state=state)
    assert token_data.claims["sub"] == subject
    assert token_data.claims["name"] == claims["name"]
    assert token_data.claims["website"] == claims["website"]
    assert token_data.claims["email"] == claims["email"]
    assert token_data.claims["address"]["formatted"] == claims["address"]["formatted"]  # type: ignore
    assert token_data.claims["phone"] == claims["phone"]
    assert token_data.scope == "openid profile email address phone"

    user_info = client.fetch_userinfo(token=token_data.access_token)
    assert user_info["sub"] == subject
    assert user_info["name"] == claims["name"]
    assert user_info["website"] == claims["website"]
    assert user_info["email"] == claims["email"]
    assert user_info["address"]["formatted"] == claims["address"]["formatted"]
    assert user_info["phone"] == claims["phone"]


def test_auth_denied(oidc_server: str):
    state = faker.password()

    client = _fake_client(oidc_server)

    response = httpx.post(
        client.authorization_url(state=faker.password()),
        data={"action": "deny"},
    )

    with pytest.raises(AuthorizationError, match=r"access_denied"):
        client.fetch_token(response.headers["location"], state=state)


@use_provider_config(require_client_registration=True)
def test_client_not_registered(oidc_server: str):
    state = faker.password()

    client = _fake_client(oidc_server)

    response = httpx.post(
        client.authorization_url(state=state),
        data={"sub": faker.email()},
    )

    assert response.status_code == HTTPStatus.BAD_REQUEST
    assert "Error: invalid_client" in response.text
    assert "The client does not exist on this server" in response.text


def test_wrong_client_secret(oidc_server: str):
    state = faker.password()
    redirect_uri = faker.uri(schemes=["https"])

    client = OidcClient.register(oidc_server, redirect_uri=redirect_uri)

    # Create a second client with the same ID but different secret
    client = OidcClient(
        issuer=oidc_server,
        id=client.id,
        secret="foobar",
        redirect_uri=redirect_uri,
    )

    response = httpx.post(
        client.authorization_url(state=state),
        data={"sub": faker.email()},
    )

    with pytest.raises(OAuthError, match="invalid_client: "):
        client.fetch_token(response.headers["location"], state=state)


@pytest.mark.parametrize(
    "auth_method",
    [
        "client_secret_basic",
        "client_secret_post",
    ],
)
def test_client_auth_methods(oidc_server: str, auth_method: str):
    subject = faker.email()
    state = faker.password()

    client = _fake_client(oidc_server, auth_method=auth_method)
    auth_url = client.authorization_url(state=state)
    response = httpx.post(auth_url, data={"sub": subject})

    token_data = client.fetch_token(response.headers["location"], state)
    assert token_data.claims["sub"] == subject

    userinfo = client.fetch_userinfo(token=token_data.access_token)
    assert userinfo["sub"] == subject


def test_auth_methods_not_supported_for_client(oidc_server: str):
    state = faker.password()

    redirect_uri = faker.uri()
    client = OidcClient.register(
        oidc_server, redirect_uri=redirect_uri, auth_method="client_secret_basic"
    )
    client = OidcClient(
        id=client.id,
        redirect_uri=redirect_uri,
        auth_method="client_secret_post",
        secret=client.secret,
        issuer=oidc_server,
    )
    auth_url = client.authorization_url(state=state)
    response = httpx.post(auth_url, data={"sub": faker.email()})
    with pytest.raises(OAuthError, match="invalid_client: "):
        client.fetch_token(response.headers["location"], state=state)


@use_provider_config(require_nonce=True)
def test_nonce_required_error(oidc_server: str):
    state = faker.password()

    client = _fake_client(oidc_server)
    auth_url = client.authorization_url(state=state)
    token_data = httpx.post(auth_url, data={"sub": faker.email()})
    with pytest.raises(
        AuthorizationError,
        match=re.compile(
            r"Authorization failed: invalid_request: Missing ['\"]nonce['\"] in request"
        ),
    ):
        client.fetch_token(token_data.headers["location"], state=state)

    nonce = faker.password()
    auth_url = client.authorization_url(state=state, nonce=nonce)
    token_data = httpx.post(auth_url, data={"sub": faker.email()})
    token_data = client.fetch_token(token_data.headers["location"], state=state)
    assert token_data.claims["nonce"] == nonce


def test_no_openid_scope(oidc_server: str):
    state = faker.password()

    client = _fake_client(oidc_server, scope="foo bar")

    response = httpx.post(
        client.authorization_url(state=state),
        data={"sub": faker.email()},
    )

    with pytest.raises(
        AuthorizationServerError, match="missing id_token from token endpoint response"
    ):
        client.fetch_token(response.headers["location"], state)


def test_no_email_scope(oidc_server: str):
    state = faker.password()

    client = OidcClient.register(
        oidc_server,
        scope="openid",
        redirect_uri=faker.uri(schemes=["https"]),
    )

    response = httpx.post(
        client.authorization_url(state=state),
        data={"sub": faker.email()},
    )

    token_data = client.fetch_token(response.headers["location"], state)
    assert token_data.scope == "openid"
    assert "email" not in token_data.claims


def test_reduced_authorization_scope(oidc_server: str):
    state = faker.password()

    client = OidcClient.register(
        oidc_server,
        scope="openid other",
        redirect_uri=faker.uri(schemes=["https"]),
    )

    response = httpx.post(
        client.authorization_url(state=state, scope="openid other notallowed"),
        data={"sub": faker.email()},
    )

    token_data = client.fetch_token(response.headers["location"], state)
    assert token_data.scope == "openid other"


@use_provider_config(access_token_max_age=timedelta(minutes=111))
def test_token_expiry(oidc_server: str):
    state = faker.password()

    client = OidcClient.register(oidc_server, redirect_uri=faker.uri(schemes=["https"]))

    response = httpx.post(
        client.authorization_url(state=state),
        data={"sub": faker.email()},
    )
    assert response.status_code == 302
    token_data = client.fetch_token(response.headers["location"], state=state)

    assert token_data.claims["exp"] - token_data.claims["iat"] == 111 * 60  # type: ignore
    assert token_data.expires_in == 111 * 60


@use_provider_config(issue_refresh_token=False)
def test_no_refresh_token(oidc_server: str):
    state = faker.password()

    client = _fake_client(oidc_server)

    response = httpx.post(
        client.authorization_url(state=state),
        data={"sub": faker.email()},
    )
    assert response.status_code == 302
    token_data = client.fetch_token(response.headers["location"], state=state)

    assert token_data.refresh_token is None


def test_refresh_token(oidc_server: str):
    state = faker.password()

    client = _fake_client(oidc_server)

    response = httpx.post(
        client.authorization_url(state=state),
        data={"sub": faker.email()},
    )

    token_data = client.fetch_token(response.headers["location"], state=state)

    assert token_data.refresh_token is not None
    refresh_token_data = client.refresh_token(refresh_token=token_data.refresh_token)

    # Using a refresh token revokes the old access token
    with pytest.raises(httpx.HTTPStatusError) as e:
        client.fetch_userinfo(token=token_data.access_token)
    assert e.value.response.json()["error"] == "access_denied"

    client.fetch_userinfo(token=refresh_token_data.access_token)


def test_revoke_tokens(oidc_server: str):
    state = faker.password()
    sub = faker.email()
    client = _fake_client(oidc_server)

    auth_response = httpx.post(
        client.authorization_url(state=state),
        data={"sub": sub},
    )
    token_data = client.fetch_token(auth_response.headers["location"], state=state)

    httpx.post(f"{oidc_server}users/{sub}/revoke-tokens").raise_for_status()

    with pytest.raises(httpx.HTTPStatusError) as e:
        client.fetch_userinfo(token=token_data.access_token)
    assert e.value.response.json()["error"] == "access_denied"

    assert token_data.refresh_token
    with pytest.raises(OAuthError, match="invalid_grant: invalid refresh token"):
        client.refresh_token(faker.password())


def test_isssue_invalid_grant_type(client: flask.testing.FlaskClient):
    response = client.post("/oauth2/token", data={"grant_type": "foo"})
    assert response.json == {"error": "unsupported_grant_type"}


@use_provider_config(access_token_max_age=timedelta(minutes=111))
def test_userinfo_expired_token(oidc_server: str):
    with freeze_time(faker.date(), tick=True) as frozen_datetime:
        state = faker.password()
        client = _fake_client(oidc_server)
        token_data = httpx.post(
            client.authorization_url(state=state),
            data={"sub": faker.email()},
        )

        token_data = client.fetch_token(token_data.headers["location"], state=state)
        frozen_datetime.tick(timedelta(minutes=112))
        with pytest.raises(httpx.HTTPStatusError) as e:
            client.fetch_userinfo(token=token_data.access_token)

        response = e.value.response.json()
        assert response["error"] == "invalid_token"


def _fake_client(
    issuer: str,
    *,
    scope: str = OidcClient.DEFAULT_SCOPE,
    auth_method: str = OidcClient.DEFAULT_AUTH_METHOD,
):
    return OidcClient(
        id=str(faker.uuid4()),
        secret=faker.password(),
        redirect_uri=faker.uri(schemes=["https"]),
        issuer=issuer,
        scope=scope,
        auth_method=auth_method,
    )



# pyright: reportUnknownMemberType=none
"""Test OIDC login of a Flask app using
[Flask-OIDC](https://flask-oidc.readthedocs.io/en/stable/).
"""

import logging
from pathlib import Path
from urllib.parse import quote

import flask
import flask.testing
import httpx
import pytest
from flask_oidc import OpenIDConnect
from playwright.sync_api import Page, expect
from pytest_flask.live_server import LiveServer

import oidc_provider_mock


def build_app(oidc_provider_url: str):
    app = flask.Flask(__name__)
    app.config.update({
        "OIDC_CLIENT_SECRETS": Path(__file__).parent / "flask_oidc_client_secrets.json",
        "OIDC_INTROSPECTION_AUTH_METHOD": "client_secret_basic",
        "OIDC_SCOPES": "openid profile email",
        "SECRET_KEY": "some secret",
        "SERVER_NAME": "localhost",  # See https://github.com/pytest-dev/pytest-flask/issues/175
    })

    @app.route("/")
    def index():
        user = flask.g.oidc_user
        if user.logged_in:
            return f"Welcome {user.profile['name']} ({user.email})"
        else:
            return "Not logged in"

    app.config["OIDC_SERVER_METADATA_URL"] = (
        f"{oidc_provider_url}/.well-known/openid-configuration"
    )
    OpenIDConnect(app)

    return app


@pytest.fixture
def app(oidc_server: str):
    logging.getLogger("oidc_provider_mock").setLevel(logging.DEBUG)
    app = build_app(oidc_server)
    return app


# SNIPPET flask_test_login


@pytest.fixture
def oidc_server():
    with oidc_provider_mock.run_server_in_thread() as server:
        yield f"http://localhost:{server.server_port}"


def test_auth_code_login(client: flask.testing.FlaskClient, oidc_server: str):
    # Add OIDC claims for the user we want to authenticate
    response = httpx.put(
        f"{oidc_server}/users/{quote('alice@example.com')}",
        json={"email": "alice@example.com", "name": "Alice", "custom": ["foo", "bar"]},
    )
    assert response.status_code == 204

    # Start login on the client and get the authorization URL
    response = client.get("/login")
    assert response.location

    # Authorize the client by POSTing to the authorization URL.
    response = httpx.post(response.location, data={"sub": "alice@example.com"})

    # Go back to the client with the authorization code
    assert response.has_redirect_location
    response = client.get(response.headers["location"], follow_redirects=True)

    # Check that we have been authenticated
    assert response.text == "Welcome Alice (alice@example.com)"


# SNIPPET END

# SNIPPET flask_test_login_playwright


def test_auth_code_login_playwright(
    live_server: LiveServer, oidc_server: str, page: Page
):
    # Let the OIDC provider know about the user’s email and name
    response = httpx.put(
        f"{oidc_server}/users/{quote('alice@example.com')}",
        json={"email": "alice@example.com", "name": "Alice"},
    )
    assert response.status_code == 204

    # Start login and be redirected to the provider
    page.goto(live_server.url("/login"))

    # Authorize with the provider
    page.get_by_placeholder("sub").fill("alice@example.com")
    page.get_by_role("button", name="Authorize").click()

    # Verify that we’re logged in
    expect(page.locator("body")).to_contain_text("Welcome Alice (alice@example.com)")


# SNIPPET END


def test_auth_denied_playwright(live_server: LiveServer, oidc_server: str, page: Page):
    """Test user denying authorization"""

    # Start login and be redirected to the provider
    page.goto(live_server.url("/login"))

    # Deny authorization
    page.get_by_role("button", name="Deny").click()

    # Verify that we’re shown an error message
    expect(page.get_by_role("heading")).to_have_text("Unauthorized")
    expect(page.locator("body")).to_contain_text(
        "access_denied: The resource owner or authorization server denied the request"
    )



import re
from http import HTTPStatus
from urllib.parse import urlencode

import flask.testing
import httpx
import pytest
from faker import Faker
from playwright.sync_api import Page, expect

from oidc_provider_mock._client_lib import OidcClient

from .conftest import use_provider_config

faker = Faker()


@pytest.mark.parametrize("method", ["GET", "POST"])
@use_provider_config(require_client_registration=True)
def test_invalid_client(client: flask.testing.FlaskClient, method: str):
    """
    Respond with 400 and error description when:

    * client_id query parameter is missing
    * client unknown
    * redirect_uri does not match the URI that was registered
    """

    query = urlencode({
        "redirect_uri": "foo",
        "response_type": "code",
    })
    response = client.open(f"/oauth2/authorize?{query}", method=method)
    assert response.status_code == 400
    assert "Error: invalid_client" in response.text
    assert "Missing &#39;client_id&#39; parameter" in response.text

    query = urlencode({
        "client_id": "UNKNOWN",
        "redirect_uri": "foo",
        "response_type": "code",
    })
    response = client.open(f"/oauth2/authorize?{query}", method=method)
    assert response.status_code == 400
    assert "Error: invalid_client" in response.text
    assert "The client does not exist on this server" in response.text

    redirect_uris = [faker.uri(schemes=["https"])]
    response = client.post(
        "/oauth2/clients",
        json={
            "redirect_uris": redirect_uris,
        },
    )
    assert response.status_code == HTTPStatus.CREATED
    oidc_client = response.json
    assert oidc_client

    query = urlencode({
        "client_id": oidc_client["client_id"],
        "redirect_uri": "foo",
        "response_type": "code",
    })
    response = client.open(f"/oauth2/authorize?{query}", method=method)
    assert response.status_code == 400
    assert "Error: invalid_client" in response.text
    assert "Redirect URI foo is not supported by client." in response.text


@pytest.mark.parametrize("method", ["GET", "POST"])
def test_missing_redirect_uri(client: flask.testing.FlaskClient, method: str):
    query = urlencode({
        "client_id": str(faker.uuid4()),
        "response_type": "code",
    })
    response = client.open(f"/oauth2/authorize?{query}", method=method)
    assert response.status_code == HTTPStatus.BAD_REQUEST
    assert response.json
    assert response.json["error"] == "invalid_request"
    assert re.match(
        r"Missing ['\"]redirect_uri['\"] in request\.",
        response.json["error_description"],
    )


def test_missing_sub_parameter(client: flask.testing.FlaskClient):
    query = urlencode({
        "client_id": str(faker.uuid4()),
        "redirect_uri": faker.uri(schemes=["https"]),
        "response_type": "code",
    })
    response = client.post(f"/oauth2/authorize?{query}")
    assert "The field is missing" in response.text


def test_authorized_users_buttons_appear_xxx(oidc_server: str, page: Page):
    """Test that authorizing 3 users creates buttons on the auth page"""

    client = OidcClient(
        id=str(faker.uuid4()),
        secret=faker.password(),
        redirect_uri=faker.uri(schemes=["https"]),
        issuer=oidc_server,
    )

    subject0 = faker.email()
    subject1 = faker.email()

    for sub in [subject0, subject1, subject0]:
        httpx.post(
            client.authorization_url(state=faker.password()),
            data={"sub": sub},
        )

    page.goto(f"{oidc_server}/oidc/login")
    page.get_by_role("button", name="Start").click()

    expect(
        page.get_by_role("heading", name="Authenticate previous users")
    ).to_be_visible()

    auth_buttons = page.get_by_role(
        "form", name="Authenticate previous users"
    ).get_by_role("button")

    expect(auth_buttons.nth(0)).to_have_accessible_name(subject0)
    expect(auth_buttons.nth(1)).to_have_accessible_name(subject1)

    auth_buttons.nth(0).click()
    expect(page.locator("body")).to_contain_text(f"You’re logged in as {subject0}")



import logging
import os
import sys
import time
import traceback
from datetime import timedelta

import click
import uvicorn

from . import app
from ._app import Config

_default_config = Config


@click.command(
    context_settings={"max_content_width": 100, "help_option_names": ["-h", "--help"]}
)
@click.option(
    "-p",
    "--port",
    help="Port the server listens on",
    default=9400,
    show_default=True,
)
@click.option(
    "-H",
    "--host",
    help="IP address to bind the server to",
    default="127.0.0.1",
    show_default=True,
)
@click.option(
    "-r",
    "--require-registration",
    help="Require clients to register before they can request authentication",
    show_default=True,
    flag_value=True,
    default=_default_config.require_client_registration,
    type=bool,
    is_flag=False,
)
@click.option(
    "-n",
    "--require-nonce",
    help="Require clients to include a nonce in the authorization request to prevent replay attacks",
    show_default=True,
    flag_value=True,
    default=_default_config.require_nonce,
    type=bool,
    is_flag=False,
)
@click.option(
    "-f",
    "--no-refresh-token",
    help="Do not issue an refresh token",
    show_default=True,
    flag_value=True,
    default=not _default_config.issue_refresh_token,
    type=bool,
    is_flag=False,
)
@click.option(
    "-e",
    "--token-max-age",
    help="Max age of access and ID tokens in seconds until they expire",
    default=_default_config.access_token_max_age.total_seconds(),
    type=int,
)
def run(
    port: int,
    host: str,
    *,
    require_registration: bool,
    require_nonce: bool,
    no_refresh_token: bool,
    token_max_age: int,
):
    """Start an OpenID Connect Provider for testing"""

    os.environ["AUTHLIB_INSECURE_TRANSPORT"] = "1"
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(
        Logfmter(
            color=not os.getenv("NO_COLOR")
            and (handler.stream.isatty() or bool(os.getenv("FORCE_COLOR")))
        )
    )
    logging.getLogger().addHandler(handler)
    logging.getLogger().setLevel(logging.INFO)

    uvicorn.run(
        app(
            require_client_registration=require_registration,
            require_nonce=require_nonce,
            issue_refresh_token=not no_refresh_token,
            access_token_max_age=timedelta(seconds=token_max_age),
        ),
        interface="wsgi",
        port=port,
        host=host,
        log_config=None,
    )


_LOG_RECORD_ATTRIBUTES = (
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "message",
    "module",
    "msecs",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "taskName",
    "thread",
    "threadName",
)


_ANSI_RESET = "\033[0m"
_ANSI_BOLD = "\033[1m"
_ANSI_RED = "\033[31m"
_ANSI_YELLOW = "\033[33m"
_ANSI_BLUE = "\033[34m"
_ANSI_WHITE = "\033[37m"


class Logfmter(logging.Formatter):
    def __init__(self, color: bool):
        self._color = color

    def format(self, record: logging.LogRecord) -> str:
        out = ""

        log_time = time.localtime(record.created)
        out += f"{time.strftime('%H:%M:%S', log_time)}.{record.msecs:03.0f}"

        out += " "
        if self._color:
            if record.levelno >= logging.ERROR:
                out += _ANSI_RED
            elif record.levelno >= logging.WARNING:
                out += _ANSI_YELLOW
                level_name = "WARN"
            elif record.levelno >= logging.INFO:
                out += _ANSI_BLUE
            else:
                out += _ANSI_WHITE

        level_name = record.levelname
        if level_name == "WARNING":
            level_name = "WARN"
        out += f"{level_name:6s}"

        if self._color:
            out += _ANSI_RESET

        out += f" {record.name}"

        if isinstance(record.msg, dict):
            data: dict[str, object] = dict(record.msg)  # pyright: ignore
        else:
            data = {}

        data.update({
            key: value
            for key, value in record.__dict__.items()
            if key not in _LOG_RECORD_ATTRIBUTES
        })

        color_message = data.pop("color_message", None)
        unformatted_message = data.pop("_msg", None)
        if self._color and isinstance(color_message, str):
            if record.args:
                out += " " + color_message % record.args
            else:
                out += " " + color_message
        elif unformatted_message:
            out += " " + str(unformatted_message)
        else:
            out += " " + record.getMessage()

        for key, value in data.items():
            out += f" {self._format_key(key)}={self._format_value(value)}"

        if record.exc_info:
            formatted_exception = "\n".join(
                traceback.format_exception(record.exc_info[1])
            )

            out += f" {self._format_key('exc_info')}={self._format_value(formatted_exception)}"

        return out

    @classmethod
    def _format_value(cls, value: object) -> str:
        if value is None:
            return ""
        elif isinstance(value, bool):
            return "true" if value else "false"

        value = str(value)

        if '"' in value:
            value = value.replace('"', '\\"')

        if "\n" in value:
            value = value.replace("\n", "\\n")

        if " " in value or "=" in value:
            value = f'"{value}"'

        return value

    def _format_key(self, key: str) -> str:
        if self._color:
            return f"{_ANSI_BOLD}{key}{_ANSI_RESET}"
        else:
            return key


if __name__ == "__main__":
    run()



from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from urllib.parse import parse_qsl, urljoin, urlparse

import httpx
import joserfc.jwk
import joserfc.jwt
import pydantic
from authlib.integrations.httpx_client import OAuth2Client


@dataclass(kw_only=True, frozen=True)
class TokenData:
    """Payload of a successful access token response.

    See https://www.rfc-editor.org/rfc/rfc6749.html#section-5.1"""

    access_token: str
    expires_in: int
    refresh_token: str | None
    claims: dict[str, object]
    scope: str | None


@dataclass(kw_only=True, frozen=True)
class RefreshTokenData:
    access_token: str
    expires_in: int
    refresh_token: str | None
    claims: dict[str, object] | None


class _OidcClaims(pydantic.BaseModel):
    """Claims defined for the OpenID JWT.

    See https://openid.net/specs/openid-connect-core-1_0.html#IDToken"""

    iss: str
    aud: str | Sequence[str]
    azp: str | None = None
    exp: int
    iat: int


class InvalidClaim(Exception):
    # Name of the invalid claim, e.g. iss, aud, etc.
    name: str

    def __init__(self, name: str, message: str) -> None:
        self.name = name
        super().__init__(message)


class OidcClient:
    DEFAULT_SCOPE = "openid email"
    DEFAULT_AUTH_METHOD = "client_secret_basic"

    _authlib_client: OAuth2Client

    def __init__(
        self,
        *,
        id: str,
        redirect_uri: str,
        secret: str,
        issuer: str,
        auth_method: str = DEFAULT_AUTH_METHOD,
        scope: str = DEFAULT_SCOPE,
    ) -> None:
        self._id = id
        self._secret = secret
        self._scope = scope
        self._issuer = issuer

        # TODO: validate response
        config = self.get_authorization_server_metadata(issuer)

        self._jwks = joserfc.jwk.KeySet.import_key_set(
            httpx.get(config["jwks_uri"]).json()
        )

        self._issuer = config["issuer"]
        self._token_endpoint_url = config["token_endpoint"]
        self._userinfo_enpoint_url = config["userinfo_endpoint"]
        self._authorization_endpoint_url = config["authorization_endpoint"]

        self._auth_method = auth_method

        self._authlib_client = OAuth2Client(
            client_id=self._id,
            client_secret=self._secret,
            token_endpoint_auth_method=auth_method,
            redirect_uri=redirect_uri,
        )

    @classmethod
    def get_authorization_server_metadata(cls, provider_url: str):
        # TODO: validate response schema
        return (
            httpx.get(
                urljoin(provider_url, ".well-known/openid-configuration"),
                follow_redirects=True,
            )
            .raise_for_status()
            .json()
        )

    @classmethod
    def register(
        cls,
        issuer: str,
        redirect_uri: str,
        scope: str = DEFAULT_SCOPE,
        auth_method: str = "client_secret_basic",
    ):
        """Register a client with the OpenID provider and instantiate it."""

        config = cls.get_authorization_server_metadata(issuer)

        # TODO: handle
        if endpoint := config.get("registration_endpoint"):
            content = (
                httpx.post(
                    endpoint,
                    json={
                        "redirect_uris": [redirect_uri],
                        "token_endpoint_auth_method": auth_method,
                        "scope": scope,
                    },
                )
                .raise_for_status()
                .json()
            )

        else:
            # TODO: Dedicated error class
            raise Exception(
                "Authorization server does not advertise registration endpoint"
            )

        return cls(
            id=content["client_id"],
            redirect_uri=redirect_uri,
            scope=scope,
            issuer=issuer,
            secret=content["client_secret"],
        )

    @property
    def secret(self) -> str:
        return self._secret

    @property
    def id(self) -> str:
        return self._id

    def authorization_url(
        self,
        *,
        state: str,
        scope: str | None = None,
        response_type: str = "code",
        nonce: str | None = None,
    ) -> str:
        if scope is None:
            scope = self._scope
        extra = {
            "scope": scope,
            "response_type": response_type,
        }
        if nonce is not None:
            extra["nonce"] = nonce

        url, _state = self._authlib_client.create_authorization_url(  # pyright: ignore[reportUnknownVariableType,reportUnknownMemberType]
            self._authorization_endpoint_url,
            state,
            code_verifier=None,
            **extra,
        )
        assert isinstance(url, str)
        return url

    def fetch_token(
        self,
        auth_response_location: str,
        state: str,
    ) -> TokenData:
        # TODO: add nonce argument and check it
        """Parse authorization endpoint response embedded in the redirect location
        and fetches the token.


        :raises AuthorizationError: if authorization was unsuccessful.
        """
        query = urlparse(auth_response_location).query
        params = dict(parse_qsl(query))

        if error := params.get("error"):
            raise AuthorizationError(error, params.get("error_description"))

        if "state" not in params:
            raise AuthorizationServerError(
                "state parameter missing from authorization response"
            )
        if params["state"] != state:
            raise AuthorizationServerError(
                "state parameter in authorization_response does not match expected value"
            )

        # TODO: wrap authlib_integrations.base_client.OAuthError
        authlib_token = self._authlib_client.fetch_token(  # pyright: ignore[reportUnknownVariableType,reportUnknownMemberType]
            self._token_endpoint_url,
            state=state,
            authorization_response=auth_response_location,
        )

        try:
            response = _TokenResponse.model_validate(authlib_token)
        except pydantic.ValidationError as e:
            # TODO: include validation error information
            raise AuthorizationServerError("invalid token endpoint response") from e

        if response.id_token is None:
            raise AuthorizationServerError(
                "missing id_token from token endpoint response"
            )

        claims = self._decode_and_verify_id_token(response.id_token)

        return TokenData(
            access_token=response.access_token,
            expires_in=response.expires_in,
            claims=claims,
            refresh_token=response.refresh_token,
            scope=response.scope,
        )

    def fetch_userinfo(self, token: str):
        # TODO: validate response schema
        return (
            httpx.get(
                self._userinfo_enpoint_url, headers={"authorization": f"bearer {token}"}
            )
            .raise_for_status()
            .json()
        )

    def refresh_token(self, refresh_token: str) -> RefreshTokenData:
        """Fetch a fresh access token using the refresh token as a grant."""

        authlib_token = self._authlib_client.fetch_token(  # pyright: ignore[reportUnknownVariableType,reportUnknownMemberType]
            self._token_endpoint_url,
            refresh_token=refresh_token,
            grant_type="refresh_token",
        )

        try:
            response = _TokenResponse.model_validate(authlib_token)
        except pydantic.ValidationError as e:
            # TODO: include validation error information
            raise AuthorizationServerError("invalid token endpoint response") from e

        if response.id_token:
            claims = self._decode_and_verify_id_token(response.id_token)
        else:
            claims = None

        return RefreshTokenData(
            access_token=response.access_token,
            expires_in=response.expires_in,
            claims=claims,
            refresh_token=response.refresh_token,
        )

    def _decode_and_verify_id_token(self, id_token: str) -> dict[str, object]:
        # See https://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation

        # 1. decode and verify signature
        token = joserfc.jwt.decode(id_token, self._jwks)
        # TODO: wrap error
        claims = _OidcClaims.model_validate(token.claims)

        # 2. iss
        if claims.iss != self._issuer:
            raise InvalidClaim("iss", f"expected {self._issuer} got {claims.iss}")

        # 3. aud
        if isinstance(claims.aud, str):
            if claims.aud != self._id:
                raise InvalidClaim("aud", f"expected {self._id} got {claims.aud}")
        else:
            aud = set(claims.aud)
            if self._id not in aud:
                raise InvalidClaim("aud", f"client ID {self._id} not included")
            untrusted = aud - {self._id}
            if untrusted:
                raise InvalidClaim(
                    "aud", f"includes untrusted audiences {', '.join(untrusted)}"
                )

        # 4. azp extension not implemented

        # 5. azp
        if claims.azp is not None and claims.azp != self._id:
            raise InvalidClaim("azp", f"expected {self._id} got {claims.azp}")

        # 6. TLS verification skipped, we’re using the signature
        # TODO: 7. Implement alg check
        # TODO: 8. Client secret check of HMAC

        now = datetime.now(tz=timezone.utc)

        # 9. exp
        exp = datetime.fromtimestamp(claims.exp, tz=timezone.utc)
        if now > exp + timedelta(seconds=5):
            raise ValueError("exp")

        # 10. iat
        iat = datetime.fromtimestamp(claims.iat, tz=timezone.utc)
        if now < iat - timedelta(hours=1):
            raise ValueError("iat")

        # 11. TODO nonce
        # 12. acr extension not implemented

        return token.claims


class AuthorizationServerError(Exception):
    """The authorization server sent an invalid response.

    For example, the server did not return an access token from the token endpoint
    response.
    """

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class AuthorizationError(Exception):
    """The authorization server responded with an error to the authorization request.

    See [OAuth2.0 Authorization Error
    Response](https://www.rfc-editor.org/rfc/rfc6749.html#section-4.1.2.1).
    """

    def __init__(self, error: str, description: str | None = None) -> None:
        self.error = error
        self.description = description

        msg = f"Authorization failed: {error}"
        if description:
            msg = f"{msg}: {description}"

        super().__init__(msg)


class _TokenResponse(pydantic.BaseModel):
    """Response body for successful requests to the token endpoint.

    See https://www.rfc-editor.org/rfc/rfc6749.html#section-5 and
    https://openid.net/specs/openid-connect-core-1_0.html#TokenResponse
    """

    access_token: str
    expires_in: int
    refresh_token: str | None = None
    id_token: str | None = None
    scope: str | None = None
